% cilog axiomatization of probability over workflow events.  This uses
% DCGs to represent the grammar (but uses Prolog syntax, not DCG
% syntax).

% In the DCG syntac used here, the grammar rules end in "D0,D1)",
% where D1 is a tail of D0, which means that the elements of D0 before
% D1 form the syntactic category of the predicate. Extra arguments
% will appear before D0.

% initally this example will be simplified so that
%   -- message can only have one attachment (CILog handles lists fine)
%   -- we are not modelling the contents of files (editing changes the contents)
%   -- we only track one file (they can only attach the last file)

% The atomic events are
%   receive(Msg,Sender,A) where A is an attachment (A=none for no attachment)
%   save(A,F) save attachment A as file F
%   edit(F) edit the file F      -- note that we are not modelling the contents
%   save(F) save the file F
%   saveas(F1,F2) save file F1 as F2
%   compose(Msg) compose a message
%   attach(F,Msg) attach file F to Msg
%   send(Msg) send message


% a workflow consists of a receive phase, an edit phase, and a send phase

workflow(D0,D3) <-
   receive_phase(F0,D0,D1) &
   edit_phase(F0,F1,D1,D2) &
   send_phase(F1,D2,D3).

% receive_phase(F,D0,D2) means F is the file that has been received
receive_phase(F,D0,D2) <-
    receive_msg(A,D0,D1) &
    save_msg(A,F,D1,D2).

% receive_msg(A,D0,D1) attachment A was received
receive_msg(A,[receive(Msg,Sender,A)|R],R) <-
    ~ receive_lost([receive(Msg,Sender,A)|R]).
receive_msg(A,R,R) <- receive_lost(R).

prob receive_lost(R): 0.01.

% save_msg(A,F,D0,D1) attachment A was saved as file F
save_msg(A,F,[save(A,F)|R],R) <-
  A \= none &
  att_saved([save(A,F)|R]).

same_msg(none,none,R,R).

save_msg(A,none,R,R) <-
   att_not_saved(R).

save_msg(A,F,R,R) <-
   F \= none &
   att_save_lost(R).

prob att_saved(R):0.8,       % attachment was explcility saved
     att_not_saved(R):0.15,  % attachment was not saved
     att_save_lost(R):0.05.  % attachment was saved but save wasn't recorded

% edit_phase(F0,F1,D1,D2)
edit_phase(F,F,D,D) <-
   no_editing(D).
edit_phase(F0,F2,D0,D2) <-
   ~ no_editing(D0) &
   edit(F0,F1,D0,D1) &
   more_editing(F1,F2,D1,D2).

prob no_editing(D):0.3.   % the probability that there is no editing

% an edit consists of an edit action followed by repeated saves
% perhaps ended by a saveas

edit(F0,F1,[edit(F0)|D1],D2) <-
   saving_phase(F0,F1,D1,D2).

saving_phase(F0,F1,[saveas(F0,F1)|D],D) <-
   end_saving_with_saveas([saveas(F0,F1)|D]).
saving_phase(F0,F2,[save(F0)|D1],D2) <-
   keep_saving([save(F0)|D1]) &
   saving_phase(F0,F2,D1,D2).
saving_phase(F0,F0,D,D) <-
   stop_saving(D).

prob  end_saving_with_saveas(D):0.3,
      keep_saving(D):0.5,
      stop_saving(D):0.2.

more_editing(F,F,D,D) <-
   ~reedit(D).

more_editing(F0,F2,D0,D2) <-
   reedit(D0) &
   edit(F0,F1,D0,D1) &
   more_editing(F1,F2,D1,D2).

prob reedit(D):0.3.   % the probability that they will edit again

%%% Send Phase %%%
send_phase(F,D,D) <-
   no_sending(D).
send_phase(F,[compose(Msg)|D0],D2) <-
   ~ no_sending([compose(Msg)|D0]) &
   optional_attach(F,Msg,Att,D0,D1) &
   send_msg(Msg,D0,D1).

prob no_sending(D): 0.2.

optional_attach(F,Msg,F,[attach(F,Msg)|D],D) <-
       do_attach([attach(F,Msg)|D]).

optional_attach(F,Msg,none,D,D) <-
       ~do_attach(D).

prob do_attach(D):0.7.

send_msg(Msg,D0,D0) <-
   send_lost(D0).

send_smg(Msg,[send(Msg)|D],D) <-
   ~send_lost([send(Msg)|D]).

prob send_lost(D): 0.02.

%%%% Example Queries %%%
% observe workflow([receive(m1,tom,att1), save(att1,f1),compose(m2),attach(f1,m2),send(m2)],[]).